# functions for plotting the dipole topography (packages 'akima' and 'fields' are required)
#------------------------------------------------------------------------------------------


circle <- function(loc, rad, col= "white") {
  ang <- seq(0, 2 * pi, length = 100)
  lines(loc[1] + rad * cos(ang), loc[2] + rad * sin(ang), col = col)
}

draw.orientation <- function(x, y, theta, length.line = 4, color = "black", line.width = 1){
  # we rotate 90 degrees in case we want a line that mimic the electrical current
  x.arrow.top <- x+cos(theta)*length.line
  y.arrow.top <- y+sin(theta)*length.line
  x.tail <- x-cos(theta)*length.line
  y.tail <- y-sin(theta)*length.line
  lines(c(x,x.arrow.top), c(y,y.arrow.top), lwd = line.width, col = color)   
  lines(c(x,x.tail), c(y,y.tail), lwd = line.width, col = color)
}

plotMEG.brainImage <- function(x, coords, sensordots =F, sensorlabels = F, barplot=F, cex = 1,  writetitle = NA, cols, breaks) {
  ind <- rep(T,length(x))
  if (sum(is.na(x)) > 0 ) ind[which(is.na(x))] <- F 
  grid.plot <- as.list(interp(x = coords[ind,1], y = coords[ind,2], z = x[ind], 
                              xo = seq(min(coords[ind,1]), max(coords[ind,1]), length = 150), 
                              yo = seq(min(coords[ind,2]), max(coords[ind,2]), length = 150), extrap = F))
  if (barplot) image.plot(grid.plot, axes = F, col = cols, breaks = breaks)
  else image(grid.plot, axes = F, col = cols, breaks = breaks)
  if (sensorlabels) text(coords[ind,1], coords[ind,2], (1:length(x))[ind], cex = cex)
  if (sensordots) points(coords[,1], coords[,2]) 
  if(!is.na(writetitle)) title(writetitle)
}



# functions to fit the dipole model (package 'spam' is required)
#------------------------------------------------------------------------------------------

# selecting lambda given desired edf 
lambda.edf.table <- function(timescale, coords, edf.target, x.dipole, y.dipole, orientation, h, r){
  
  time.length <- length(timescale)
  
  # build the spatial smooth for the two poles
  x.pole1 <- x.dipole + r*cos(orientation)
  y.pole1 <- y.dipole + r*sin(orientation)
  x.pole2 <- x.dipole - r*cos(orientation)
  y.pole2 <- y.dipole - r*sin(orientation)
  dst.from.pole1 <- sqrt((coords[,1]-x.pole1)^2 + (coords[,2]-y.pole1)^2)
  dst.from.pole2 <- sqrt((coords[,1]-x.pole2)^2 + (coords[,2]-y.pole2)^2)
  p1 <- exp(-0.5 * dst.from.pole1^2 / (h^2))                 
  p2 <- exp(-0.5 * dst.from.pole2^2 / (h^2))             
  
  # build XtX:
  XtX_topleft <- diag(c(p1%*%p1), time.length)
  XtX_topright <- XtX_bottomleft  <- diag(c(p1 %*% p2), time.length)
  XtX_bottomright <- diag(c(p2%*%p2), time.length)
  XtX <- rbind(cbind(XtX_topleft, XtX_topright), cbind(XtX_bottomleft, XtX_bottomright))
  Diff <- diff(diag(time.length),difference=2)
  
  lambda.edf <- select.lambda(XtX, Diff, p1, p2, edf.target, method = "df", print.trace = F)
  return(list(lambda=lambda.edf$lambda, edf=lambda.edf$edf))
}


# calculate the edf corresponding to a given value for lambda1 and lambda2
lambda.df <- function(lambda, XtX, Diff, p1, p2, print.trace = F){
  lambda1 <- lambda2 <- lambda
  length.time <- dim(Diff)[2]
  Pen <- diag(c(lambda1,lambda2),2) %x% (t(Diff)%*%Diff)
  X1 <- solve(XtX + Pen)  # symmetric matrix
  
  #  exploit the block structure of X1 (composed by 4 diagonal block matrices)
  X1_topleft <- X1[1:length.time, 1:length.time]
  X1_topright <- X1[1:length.time, (length.time+1):(length.time*2)]
  X1_bottomleft <- X1[(length.time+1):(length.time*2), 1:length.time]
  X1_bottomright <- X1[(length.time+1):(length.time*2), (length.time+1):(length.time*2)]
  
  # compute edf as the trace of the influence matrix A
  edf <- sum(diag(X1_topleft) %x% (p1^2)) + sum(diag(X1_topright) %x% (p1*p2)) + sum(diag(X1_bottomright) %x% (p2^2)) + sum(diag(X1_bottomleft) %x% (p1*p2))
  # edf <- sum(diag( X1 %*% t(X) %*% X)) # but this is slower and need also the (not cheap) construction of the design matrix X
  
  if(print.trace) print(c(lambda, edf))
  edf 
}


select.lambda <- function(XtX, Diff, p1, p2, df.gamma.smooth, method = "df", print.trace = F) {
  if (method == "df") {
    
    # find lower limit for lambda
    lambda <- 1
    while (lambda.df(lambda, XtX, Diff, p1, p2) <= df.gamma.smooth) lambda <- lambda / 10
    lower  <- lambda
    
    # find upper limit for lambda
    lambda <- 1
    while (lambda.df(lambda, XtX, Diff, p1, p2) >= df.gamma.smooth) lambda <- lambda * 10
    upper  <- lambda
    
    lambda.crit <- function(lambda, XtX, Diff, p1, p2, df.gamma.smooth)  
      lambda.df(lambda, XtX, Diff, p1, p2) - df.gamma.smooth
    result <- uniroot(lambda.crit, interval = c(lower, upper), XtX, Diff, p1, p2, df.gamma.smooth, tol= 1e-1)
    lambda <- result$root
  }
  return(list(lambda=lambda, edf=lambda.df(lambda, XtX, Diff, p1, p2)))
}

fit.dipole.spam <- function(y, timescale, coords, x.dipole, y.dipole, orientation, h, r, lambda1=100, lambda2=100, prefixed.lambda, df.gamma){
   x <- timescale
   ind.no.miss <- !( as.logical( apply(y[,],1,function(y) all(is.na(y))) ) )

   # build the spatial smooth for the two poles
   x.pole1 <- x.dipole + r*cos(orientation)
   y.pole1 <- y.dipole + r*sin(orientation)
   x.pole2 <- x.dipole - r*cos(orientation)
   y.pole2 <- y.dipole - r*sin(orientation)
   dst.from.pole1 <- sqrt((coords[,1]-x.pole1)^2 + (coords[,2]-y.pole1)^2)
   dst.from.pole2 <- sqrt((coords[,1]-x.pole2)^2 + (coords[,2]-y.pole2)^2)
   p1 <- exp(-0.5 * dst.from.pole1^2 / (h^2))                 
   p2 <- exp(-0.5 * dst.from.pole2^2 / (h^2))               

   # build XtX:
   XtX_topleft <- diag(c(p1[ind.no.miss]%*%p1[ind.no.miss]), length(x))
   XtX_topright <- XtX_bottomleft  <- diag(c(p1[ind.no.miss] %*% p2[ind.no.miss]), length(x))
   XtX_bottomright <- diag(c(p2[ind.no.miss]%*%p2[ind.no.miss]), length(x))
   XtX <- rbind(cbind(XtX_topleft, XtX_topright), cbind(XtX_bottomleft, XtX_bottomright))
 
   # build Xty:  (note: the y of "Xty" is meant to be the data vector "yvec")
   Xty <- rbind( t(p1[ind.no.miss]%*%y[ind.no.miss,]), t(p2[ind.no.miss]%*%y[ind.no.miss,]) )  

   # build the penalty matrix:
   D <- diff(diag(ncol(XtX)/2),difference=2)
   if(prefixed.lambda){
      lambda1 <- lambda1
      lambda2 <- lambda2
      edf <- lambda.df(lambda1, XtX, D, p1, p2)
   } else {
      lambda.edf <- select.lambda(XtX, D, p1, p2, df.gamma, method = "df", print.trace = F)
      lambda1 <- lambda2 <- lambda.edf$lambda
      edf <- lambda.edf$edf
   }
   Pen <- diag(c(lambda1,lambda2),2) %x% (t(D)%*%D)
   
   # compute the penalized regression coefficients "g"
   U <- chol.spam(as.spam(XtX + Pen))   
   g <- backsolve.spam(U, forwardsolve.spam(U, c(Xty)))  

   # compute the fitted values:
   yvec.hat <- as.numeric(g[1:length(x)] %x% p1 + g[(length(x)+1):(length(x)*2)] %x% p2)
   
   # compute the residual sum of squares
   rss <- sum((y[ind.no.miss,] - matrix(yvec.hat,ncol=length(x))[ind.no.miss,])^2)
   return(list(y.hat=matrix(yvec.hat,ncol=length(x)), g=g, p1=p1, p2=p2, XtX=XtX, Pen=Pen, rss=rss, edf=edf, lambda1=lambda1, lambda2=lambda2))
}

# objective function to usde inside optim()
# it returns the residual sum of squares
fun.obj.spam <- function(theta, y, timescale, coords, lambda1, lambda2, prefixed.lambda, df.gamma){
   x.dipole <- theta[1]
   y.dipole <- theta[2]
   orientation <- theta[3]
   h <- theta[4]
   r <- theta[5]
   fit.dipole.spam(y=y, timescale=timescale, coords=coords, x.dipole=x.dipole, y.dipole=y.dipole, 
             orientation=orientation, h=h, r=r, lambda1=lambda1, lambda2=lambda2, prefixed.lambda=prefixed.lambda, df.gamma=df.gamma)$rss
}
compute.dst <- function(to.xy, from.xy)      sqrt( (to.xy[1] - from.xy[,1])^2 + (to.xy[2] - from.xy[,2])^2 )


dipolefitMEG <- function(y, coords, pre.stimulus=T, edf.target=60, df.mean.gamma=7, df.amplitude=3, df.phase=3) {
   
   # check format of the data 
   if(dim(y)[1] != dim(coords)[1]) y <- t(y)
   tim <- seq(-0.5,0.5,length.out=(dim(y)[2]))
   if (pre.stimulus) ind.signal <- c(1:128) else ind.signal <- c(129:256)
   timescale <- tim[ind.signal]
   y <- y[, ind.signal]

   # pre-settings
   dst2D <- compute.dst(coords[1,], coords[,])
   min.dst <- min(dst2D[-1])
   max.dst <- max(dst2D)
   ave.dst <- mean(dst2D)
   hs <- seq( min.dst, ave.dst, length.out=3) # levels for pole size
   os <- c( 0, pi/4, pi/2, (3/4)*pi)          # levels for orientation
   rs <- c(1,1.5)                             # levels for pole distance (recommended to be not greater than 2 times the pole size)
   
   # define the spatial searching area
   searching.grid.x <- seq(coords[1,1]-(max.dst/2), coords[1,1]+(max.dst/2), length.out=5)
   searching.grid.y <- seq(coords[1,2]-(max.dst/2), coords[1,2]+(max.dst/2), length.out=5)
   coords.searching.area <- expand.grid(searching.grid.x, searching.grid.y)
   ind.even <- !as.logical(1:length(searching.grid.x) %% 2 )
   ind.odd <- as.logical(1:length(searching.grid.x) %% 2 )
   ind.searching.area.zigzag <- ind.odd
   is.odd <- T
   while (length(ind.searching.area.zigzag) < dim(coords.searching.area)[1]) {
      if (is.odd) {
         ind.searching.area.zigzag <- c(ind.searching.area.zigzag, ind.even)
         is.odd <- F
      } else {
         ind.searching.area.zigzag <- c(ind.searching.area.zigzag, ind.odd)
         is.odd <- T
      }
   }
   searching.area <- which(ind.searching.area.zigzag)

   # pre-computation of lambda in the different setting levels (hs, rs)
   lambda.table <- array(NA, dim=c(length(hs),length(rs)))
   for(ind.h in 1:length(hs)){
      for(ind.r in 1:length(rs)){
         lambda.edf.calc <- lambda.edf.table(timescale, coords=coords, edf.target=edf.target, x.dipole=coords[1,1], y.dipole=coords[1,2], orientation=0, h=hs[ind.h], r=hs[ind.h]*rs[ind.r])
         lambda.table[ind.h,ind.r] <- lambda.edf.calc$lambda
      }
   }
   print("LAMBDA TABLE computed...")
   
   # step 1, PRE-SEARCHING ALGORITHM (using a precomputed "df.table" for a specific value of edf.target)
   ind.no.miss <- !( as.logical( apply(y[,],1,function(y) all(is.na(y))) ) )
   hs.hat <- hs
   betas.hat <- os
   rs.hat <- rs
   rss.location.beta <- array(NA, dim=c(length(y[,1]), length(betas.hat), length(hs.hat), length(rs.hat)))
   for(j in 1:length(betas.hat)){  
      for(k in 1:length(hs.hat)){
        for(z in 1:length(rs.hat)) {
           lambda1 <- lambda2 <- lambda.table[k,z]
           for(i in searching.area){
              mod <- fit.dipole.spam(y=y, timescale=timescale, coords=coords, x.dipole=coords[i,1], 
                            y.dipole=coords[i,2], orientation=betas.hat[j], h=hs.hat[k], r = hs.hat[k]*rs.hat[z], 
                            lambda1=lambda1, lambda2=lambda2, prefixed.lambda=T)
              rss.location.beta[i,j,k,z] <- mod$rss
           }
        } 
      }
   }
   min.rss <- min(rss.location.beta, na.rm=T)
   max.rss <- max(rss.location.beta,na.rm=T) 
   rss.location.beta[which(is.na(rss.location.beta))] <- max.rss + 1
   min.first.dim <- (apply(rss.location.beta, 1, FUN=function(x) min(x, na.rm=T))) 
   min.second.dim <- (apply(rss.location.beta, 2, FUN=function(x) min(x, na.rm=T)))
   min.third.dim <- (apply(rss.location.beta, 3, FUN=function(x) min(x, na.rm=T)))
   min.fourth.dim <- (apply(rss.location.beta, 4, FUN=function(x) min(x, na.rm=T)))
   ind.location.hat <- which(min.first.dim == min.rss)
   beta.hat <- betas.hat[which(min.second.dim == min.rss)]
   h.hat <- hs.hat[which(min.third.dim == min.rss)]
   rs.hat <- rs.hat[which(min.fourth.dim == min.rss)]
   theta.init <- c(coords[ind.location.hat,1], coords[ind.location.hat,2], beta.hat, h.hat, h.hat*rs.hat)
   lambda.opt <- lambda.table[which(hs.hat==h.hat), which(rs.hat==rs.hat)]
   print("PRE-SEARCH STEP completed...")
     
   # step 2, REFINE the SEARCH with optim() 
   optim.final <- optim(theta.init, fun.obj.spam, y=y, timescale=timescale, coords=coords, lambda1=lambda.opt, lambda2=lambda.opt, 
                 prefixed.lambda=TRUE, control = list(trace = FALSE))     
   print("SEARCH-REFINEMENT STEP completed...")
   
   # run again fit.dipole() with the refined estimates, in order to store the fitted values and the pole signals
   x.dipole <- optim.final$par[1]         # x coord of the estimated dipole location
   y.dipole <- optim.final$par[2]         # y coord of the estimated dipole location
   orientation <- optim.final$par[3]      # estimate of the dipole orientation
   h <- optim.final$par[4]                # estimate of the pole size (equal for both poles of the dipole)
   r <- optim.final$par[5]                # estimate of the pole distance 
   fit <- fit.dipole.spam(y=y, timescale=timescale, coords=coords, x.dipole=x.dipole, y.dipole=y.dipole, orientation=orientation, h=h, r = r, 
                             lambda1=lambda.opt, lambda2=lambda.opt, prefixed.lambda=T)
   y.pole1.hat <- fit$g[1:length(timescale)]                       
   y.pole2.hat <- fit$g[(length(timescale)+1):(length(timescale)*2)]
   
   # now we model each pole signal with the complex logarithm model and extract phase and amplitude curves
   qp.pole1 <- poleqpMEG(y.pole1.hat, timescale, df.mean.gamma=df.mean.gamma, df.alpha=df.amplitude, df.phi=df.phase)
   qp.pole2 <- poleqpMEG(y.pole2.hat, timescale, df.mean.gamma=df.mean.gamma, df.alpha=df.amplitude, df.phi=df.phase)

   # calculate fitted values
   yvec.recomposed.hat <- as.numeric(qp.pole1$y.qp.decomposition %x% fit$p1 + qp.pole2$y.qp.decomposition %x% fit$p2)
   rss.recomposed <- sum((y[ind.no.miss,] - matrix(yvec.recomposed.hat,ncol=length(timescale))[ind.no.miss,])^2)
   
   # calculate mean frequency 
   freq.mean <- (qp.pole1$freq.hat + qp.pole2$freq.hat) / 2

   # calculate out of phase information
   delta <- abs(qp.pole1$phase.hat-qp.pole2$phase.hat) %% (2 * pi) 
   out.of.phase <- apply(cbind(delta, 2*pi - delta), 1, min)
       # out.of.phase is a curve over time taking values in (0,pi): it is equal to pi when the two pole signals are completely out of phase                 
   dipole.strength <- exp(-0.5*(qp.pole1$freq.hat - qp.pole2$freq.hat)^2) * exp(-0.5*(pi-out.of.phase[-length(out.of.phase)])^2/0.5^2) 
       # dipole.strength is a curve over time expressing the strength of the dipole as a function of closeness in frequency of the two pole signals
       # and closeness to complete out of phase of the two pole signals 
                          
   return( list(timescale=timescale, y.hat=matrix(yvec.recomposed.hat,ncol=length(timescale)), 
       rss.recomposed=rss.recomposed, rss=fit$rss, edf=fit$edf, 
       gcv.score= (length(c(y))*fit$rss)/((length(c(y))-fit$edf)^2), 
       y.pole1.hat=y.pole1.hat, y.pole2.hat=y.pole2.hat, 
       loc.hat=c(x.dipole, y.dipole), orientation.hat=orientation, h.hat=h, r.hat=r, lambda.opt=lambda.opt, 
       qp.pole1=qp.pole1, qp.pole2=qp.pole2, freq.mean=freq.mean, out.of.phase=out.of.phase, dipole.strength=dipole.strength) )
}      



# functions for estimating quasi-periodic components: amplitude and phase smooth curves
#------------------------------------------------------------------------------------------   

qp.initialize.amplitude <- function(y, timescale){   
   # find zero crossing of the signal "y", denoted as "u" 
   # and calculate the log(max), i.e. an estimate of the log amplitude, over each interval between two consecutive zero crossings.
   u <- numeric()
   ind <- numeric()
   sign <- y[1]/abs(y[1])
   for(i in 2:length(y)) {
      if (y[i]/abs(y[i]) != sign) {
         ind <- c(ind,i)
         u.tmp <- (timescale[i]* abs(y[i])+ timescale[i-1]*abs(y[i-1]) ) / abs(y[i]-y[i-1]) 
         u <- c(u, u.tmp)         
         sign <- y[i]/abs(y[i])
      }
   }
   a <- numeric()
   x.a <- numeric()
   for(i in 2:length(ind))    {
      a <- c(a, log(max(abs(y[ind[i-1]:ind[i]]), na.rm=T)) )
      x.a <- c(x.a, mean(timescale[ind[i-1]:ind[i]]))            
   }
   return(list(u=u, ind=ind, alpha.x=x.a,alpha.y=a))
}

qp.initialize.phase <- function(y, timescale){
   # find zero crossing of the signal "y", denoted as "u[-1]" (first el. is timescale[1]), 
   # and their correpsondent phase values denoted as "v"
   u <- numeric()
   sign <- y[1]/abs(y[1])
   v <- numeric()
   ind <- numeric()
   first.z <- T
   for(i in 2:length(y)) {
      if (y[i]/abs(y[i]) != sign) {
         ind <- c(ind,i)
         if (first.z == T) { 
            if(mean(y[1:(i-1)]) < 0) {
               K <- 1
            } else {
               K <- 0
            }
            first.z <- F
         }
         # the zero crossing "u" is chosen to be the midpoint between timescale[i:i-1]
         u.tmp <- (timescale[i]* abs(y[i])+ timescale[i-1]*abs(y[i-1]) ) / abs(y[i]-y[i-1]) 
         u <- c(u, u.tmp)
         v <- c(v, (K + (1/2))*pi)
         K <- K + 1             # update the multiple K
         sign <- y[i]/abs(y[i]) # update the sign of the signal
      }
   }
   phi <- numeric()
   x.phi <- numeric()
   for(i in 2:length(ind))    {
      phi <- c(phi, (v[i]+v[i-1])/2 )
      x.phi <- c(x.phi, mean(timescale[ind[i-1]:ind[i]]))            
   }   
   return(list(u=u, v=v, ind=ind, phi.x=x.phi,phi.y=phi))
}


poleqpMEG <- function(y, timescale, df.mean.gamma=5, df.alpha=10, df.phi=10){
  # find zero-crossing of y(t)-mean(t), evaluate initial phase and amplitude at zero-crossings 
  # and fit a smooth p-spline for phase and amplitude over timescale
  mean.y <- sm.pam(y ~ s(timescale, xrange = range(timescale), nseg=50, df=df.mean.gamma), display="none")
  intercept.hat <- mean.y$fitted  # intercept evaluated at 'timescale'
  y <- y - intercept.hat
  phase <- qp.initialize.phase(y=y, timescale=timescale)
  amplitude <- qp.initialize.amplitude(y=y, timescale=timescale)
  alpha.init <- sm.pam(amplitude$alpha.y ~ s(amplitude$alpha.x, xrange = range(timescale), nseg=50, df=df.alpha), display="none") 
  phase.init <- sm.pam(phase$phi.y ~ s(phase$phi.x, xrange = range(timescale), nseg=50, df=df.phi), display="none")
  B <- cbind(rep(1,length(timescale)),bbase(timescale,nseg=50))   
  amplitude.hat <- c(exp(B%*%alpha.init$alpha))
  phase.hat <- c(B%*%phase.init$alpha)
  y.qp.decomposition <- intercept.hat + amplitude.hat*cos(phase.hat)
  freq.hat <- diff(phase.hat)/(diff(timescale)*2*pi) 
  return(list(y.qp.decomposition=y.qp.decomposition, intercept.hat=intercept.hat, amplitude.hat=amplitude.hat, phase.hat=phase.hat, freq.hat=freq.hat,
              alpha.coef = alpha.init$alpha, phase.coef = phase.init$alpha, mean.coef = mean.y$alpha, 
              zero.crossings = phase$u))
}





# sm-pam functions
#------------------------------------------------------------------------------------------

bbase <- function(x, xl = min(x), xr = max(x), nseg = 10, deg = 3) {
  # Construct B-spline basis
  dx <- (xr - xl) / nseg
  knots <- seq(xl - deg * dx, xr + deg * dx, by = dx)
  P <- outer(x, knots, tpower, deg)
  n <- dim(P)[2]
  D <- diff(diag(n), diff = deg + 1) / (gamma(deg + 1) * dx ^ deg)
  B <- (-1) ^ (deg + 1) * P %*% t(D)
  B
}

tpower <- function(x, t, p)
  # Truncated p-th power function
  (x - t) ^ p * (x > t)

lambda.select <- function(btb, bty, D, df, method = "df") {
  lambda.df <- function(lambda, btb, D) {
    B1   <- solve(btb + lambda * D)
    sum(diag(btb %*% B1))
  }
  if (method == "df") {
    lambda <- 1
    while (lambda.df(lambda, btb, D) <= df) lambda <- lambda / 10
    lower  <- lambda
    lambda <- 1
    while (lambda.df(lambda, btb, D) >= df) lambda <- lambda * 10
    upper  <- lambda
    lambda.crit <- function(lambda, btb, D, df)
      lambda.df(lambda, btb, D) - df
    result <- uniroot(lambda.crit, interval = c(lower, upper), btb, D, df)
    lambda <- result$root
  }
  lambda
}

sm.pam <- function(pam.formula, weights, bdeg = 3, pord = 2, ...
                   # increasing = FALSE, decreasing = FALSE, kappa = lambda * 100,
                   # fixed
) {
  
  opt <- sm.options(list(...))
  
  replace.na(opt, display, "lines")
  replace.na(opt, reference, "none")
  weights.missing <- missing(weights)
  
  terms.obj     <- terms(pam.formula, specials = "s")
  vars.inf      <- eval.parent(attr(terms.obj, "variables"))
  slabels       <- attr(terms.obj, "term.labels")
  s.ind         <- attr(terms.obj, "specials")$s
  response.ind  <- attr(terms.obj, "response")
  involved      <- attr(terms.obj, "factors")
  ylab          <- attr(terms.obj, "variables")
  ylab          <- strsplit(deparse(ylab), ",")[[1]][1]
  ylab          <- substr(ylab, 6, nchar(ylab))
  y             <- unlist(vars.inf[[response.ind]])
  nterms        <- length(slabels)
  X             <- list()
  xlabels       <- list()
  xlab          <- list()
  ndims         <- list()
  df            <- list()
  nseg          <- list()
  lambda        <- list()
  period        <- list()
  xrange        <- list()
  xmissing      <- FALSE
  
  if (any(apply(involved, 2, sum) > 3))
    stop("four-way interactions not yet implemented.")
  
  for (i in 1:nterms) {
    inv          <- which(involved[ , i] == 1)
    nvars        <- length(inv)
    X[[i]]       <- matrix(nrow = length(y), ncol = 0)
    xlabels[[i]] <- vector("character")
    xlab[[i]]    <- vector("character")
    ndims[[i]]   <- numeric()
    df[[i]]      <- numeric()
    lambda[[i]]  <- numeric()
    period[[i]]  <- numeric()
    nseg[[i]]    <- numeric()
    xrange[[i]]  <- matrix( , nrow = 0, ncol = 2)
    for (j in inv) {
      newvar       <- eval.parent(parse(text = vars.inf[[j]]$variables))
      X[[i]]       <- cbind(X[[i]], newvar)
      xlabels[[i]] <- c(xlabels[[i]], vars.inf[[j]]$variables)
      if (is.matrix(newvar)) {
        nms <- colnames(newvar)
        if (any(is.null(colnames(newvar)))) 
          nms <- paste(vars.inf[[j]]$variables, 
                       "[", 1:ncol(newvar), "]", sep = "")
      }
      else 
        nms <- vars.inf[[j]]$variables
      xlab[[i]]    <- c(xlab[[i]], nms)
      ndims.new    <- if (is.vector(newvar)) 1 else ncol(newvar)
      ndims[[i]]   <- c(ndims[[i]], ndims.new)
      df.new       <- vars.inf[[j]]$df
      if (is.na(df.new)) df.new <- switch(ndims.new, 6, 12, 18)
      df[[i]]      <- c(df[[i]], df.new)
      lambda[[i]]  <- c(lambda[[i]], vars.inf[[j]]$lambda)
      period[[i]]  <- c(period[[i]], vars.inf[[j]]$period)
      nseg[[i]]    <- c(nseg[[i]],   vars.inf[[j]]$nseg)
      xrange[[i]]  <- rbind(xrange[[i]], vars.inf[[j]]$xrange)
    }
    if (any(is.na(nseg[[i]])) | prod(nseg[[i]]) > 400)
      nseg[[i]] <- rep(switch(sum(ndims[[i]]), 100, 17, 7), sum(ndims[[i]]))
    if (any(is.na(X[[i]]))) xmissing  <- TRUE
  }
  
  if (opt$verbose > 1) tim <- proc.time()
  
  P      <- list(length = nterms)
  B      <- rep(1, length(y))
  m      <- 1
  for (i in 1:nterms) {
    mat    <- ps.matrices(X[[i]], xrange[[i]], ndims = ndims[[i]], nseg = nseg[[i]])
    P[[i]] <- mat$P
    B      <- cbind(B, mat$B)
    m      <- c(m, ncol(mat$B))
    xrange[[i]] <- mat$xrange
  }
  
  if (opt$verbose > 1) {
    cat("Timings:\nconstructing matrices", (proc.time() - tim)[1], "seconds\n")
    tim <- proc.time()
  }
  
  b.ind <- list(length = length(m))
  for (i in 1:nterms)
    b.ind[[i]] <- (cumsum(m)[i] + 1):cumsum(m)[i + 1]
  
  if (weights.missing) {
    btb   <- t(B)  %*% B
    bty   <- t(B)  %*% y
  }
  else if (is.vector(weights)) {
    btb   <- t(B * weights)  %*% B
    bty   <- t(B * weights)  %*% y
  }
  else if (is.matrix(weights)) {
    btb   <- t(B) %*% weights %*% B
    bty   <- t(B) %*% weights %*% y
  }
  else
    stop("the weights argument is inappropriate.")
  
  if (opt$verbose > 1) {
    cat("matrix products", (proc.time() - tim)[1], "seconds\n")
    tim <- proc.time()
  }
  
  # Select the smoothing parameters, if required
  lambda.df <- function(lambda, btb, P) {
    B1   <- solve(btb + lambda * P)
    sum(diag(btb %*% B1))
  }
  
  for (i in 1:nterms) {
    if (any(is.na(df[[i]]))) df[[i]] <- switch(sum(ndims[[i]]), 6, 12, 18)
    # code doesn't currently handle more than one df for terms with more than one variable.
    df[[i]] <- sum(df[[i]])
    if (any(is.na(lambda[[i]])))
      lambda[[i]] <- lambda.select(btb[b.ind[[i]], b.ind[[i]]], bty[b.ind[[i]]], P[[i]], df[[i]])
  }
  
  if (opt$verbose > 1) {
    cat("selecting smoothing parameters", (proc.time() - tim)[1], "seconds\n")
    tim <- proc.time()
  }
  
  # Fit
  Pall  <- matrix(0, nrow = ncol(B), ncol = ncol(B))
  for (i in 1:nterms)
    Pall[b.ind[[i]], b.ind[[i]]] <- lambda[[i]] * P[[i]]
  B1    <- solve(btb + Pall) 
  # btb   <- t(B[,-1])  %*% diag(rep(1, length(y))) %*% B[,-1]
  # return(list(btb = btb))
  # B1    <- solve(btb[-1, -1] + Pall[-1, -1] - lambda[[i]] * mat$cmat)
  alpha <- as.vector(B1 %*% bty)
  # alpha <- as.vector(B1 %*% bty[-1])
  mu    <- c(B %*% alpha)
  # mu    <- c(B[, -1] %*% alpha)
  # return(list(alpha = alpha, B = B[ , -1], btb = btb[-1, -1]))
  
  if (opt$verbose > 1) {
    cat("fitting", (proc.time() - tim)[1], "seconds\n")
    tim <- proc.time()
  }
  
  df.model   <- sum(diag(btb %*% B1))
  df.error   <- length(y) - sum(diag(btb %*% (t(B1) + B1 - t(B1) %*% btb %*% B1)))
  sigma      <- sqrt(sum((y - mu)^2) / df.error)
  cov.alpha  <- B1 %*% btb %*% t(B1) * sigma^2
  rss        <- sum((y - mu)^2)
  tss        <- sum((y - mean(y))^2)
  R.squared  <- 100 * (tss - rss) / tss
  
  if (opt$verbose > 1) {
    cat("summaries", (proc.time() - tim)[1], "seconds\n")
    tim <- proc.time()
  }
  
  # If there is only one term, include the mean
  if (nterms == 1) b.ind[[1]] <- c(1, b.ind[[1]])
  
  result <- list(fitted = mu, alpha = alpha, m = m, B = B, bty = bty, btb = btb, B1 = B1,
                 xlabels = xlabels, xlab = xlab, ylab = ylab, slabels = slabels,
                 lambda = lambda, ndims = ndims, y = y, X = X,
                 sigma = sigma, cov.alpha = cov.alpha, b.ind = b.ind,
                 df = df, df.model = df.model, df.error = df.error, 
                 rss = rss, R.squared = R.squared, xrange = xrange,
                 nseg = nseg, bdeg = bdeg, pam.formula = pam.formula)
  if (!weights.missing) result$weights <- weights
  class(result) <- "pam"
  
  if (nterms == 1) {
    if (opt$panel) {
      replace.na(opt, df.max, switch(ndims[[1]], 20, 50, 100))
      df.min  <- switch(ndims[[1]], 2, 4, 8) + 0.1
      df.max  <- if (!opt$panel) df[[1]] else opt$df.max
      df.min  <- if (!opt$panel) df[[1]] else df.min
      Pall    <- rbind(0, cbind(0, P[[1]]))
      llambda    <- 0
      llambda.df <- lambda.df(exp(max(llambda)), btb, Pall)
      while (min(llambda.df) >= df.min) {
        llambda    <- c(llambda, max(llambda) + log(10))
        llambda.df <- c(llambda.df, lambda.df(exp(max(llambda)), btb, Pall))
      }
      while (max(llambda.df) <= df.max) {
        llambda    <- c(llambda, min(llambda) - log(10))
        llambda.df <- c(llambda.df, lambda.df(exp(min(llambda)), btb, Pall))
      }
      df.fun <- approxfun(llambda.df, llambda)
      
      sm.pam.draw <- function(pam.panel) {
        plot(pam.panel$model, options = pam.panel$opt)
        title(pam.panel$df)
        pam.panel
      }
      sm.pam.redraw <- function(pam.panel) {
        # pam.panel$model$lambda <- lambda.select(pam.panel$model$btb, pam.panel$model$bty,
        #                                         Pall, pam.panel$df)
        pam.panel$model$lambda <- exp(pam.panel$df.fun(pam.panel$df))
        B1 <- solve(pam.panel$model$btb + pam.panel$model$lambda * pam.panel$Pall)
        pam.panel$model$alpha  <- as.vector(B1 %*% pam.panel$model$bty)
        pam.panel$model$fitted <- c(pam.panel$model$B %*% pam.panel$model$alpha)
        pam.panel$opt$se       <- pam.panel$se
        pam.panel$opt$theta    <- pam.panel$theta
        pam.panel$opt$phi      <- pam.panel$phi
        rp.tkrreplot(pam.panel, plot)
        pam.panel
      }
      opt1 <- opt
      opt1$panel <- FALSE
      pam.panel <- rp.control(model = result, opt = opt1, Pall = rbind(0, cbind(0, P[[1]])),
                              df = opt$df, df.fun = df.fun, theta = opt$theta, phi = opt$phi)
      rp.tkrplot(pam.panel, plot, sm.pam.draw, hscale = opt$hscale, vscale = opt$vscale, pos = "right")
      rp.slider(pam.panel, df, df.min, df.max, sm.pam.redraw)
      rp.checkbox(pam.panel, se, sm.pam.redraw, title = "Standard errors")
      if (ndims[[1]] == 2) {
        rp.slider(pam.panel, theta, -180, 180, sm.pam.redraw, "persp angle 1")
        rp.slider(pam.panel, phi,      0,  90, sm.pam.redraw, "persp angle 2")
      }
    }
    else if (opt$display != "none")
      plot(result, ...)
  }
  
  invisible(result)
}


lambda.select <- function(btb, bty, P, df, method = "df") {
  #     This currently uses the same lambda in all dimensions
  lambda.df <- function(lambda, btb, P) {
    B1   <- solve(btb + lambda * P)
    # print(c(lambda, sum(diag(btb %*% B1))))
    sum(diag(btb %*% B1))
  }
  if (method == "df") {
    lambda <- 1
    while (lambda.df(lambda, btb, P) <= df) lambda <- lambda / 10
    lower  <- lambda
    lambda <- 1
    while (lambda.df(lambda, btb, P) >= df) lambda <- lambda * 10
    upper  <- lambda
    lambda.crit <- function(lambda, btb, P, df)
      lambda.df(lambda, btb, P) - df
    result <- uniroot(lambda.crit, interval = c(lower, upper), btb, P, df)
    # cat("result$root", result$root, "\n")
    lambda <- result$root
  }
  lambda
}

ps.matrices <- function(x, xrange, ndims, nseg, bdeg = 3, pord = 2,
                        decompose =  TRUE) {
  
  # Compute a set of basis functions and a penalty matrix associated with x.
  # An intercept term and the main effect of any interaction terms are removed.
  
  ndimx <- ncol(x)
  if (ndimx > 3) stop("terms with more than three dimensions cannot be used.")
  n    <- nrow(x)
  
  if (missing(nseg)) nseg <- rep(switch(ndimx, 100, 17, 7), ndimx)
  if (missing(xrange) | any(is.na(xrange))) {
    xrange <- cbind(apply(x, 2, min), apply(x, 2, max))
    xrange <- t(apply(xrange, 1, function(x) c(x[1] - 0.05 * diff(x), x[2] + 0.05 * diff(x))))
  }
  
  # Compute B-spline basis
  
  b <- list(length = ndimx)
  m <- vector(length = ndimx)
  for (i in 1:ndimx) {
    b[[i]] <- bbase(x[,i], xl = xrange[i , 1], xr = xrange[i, 2], nseg = nseg[i], 
                    deg = bdeg)
    m[i]   <- ncol(b[[i]])
  }
  B <- b[[1]]
  if (ndimx > 1)
    B <- t(apply(cbind(b[[1]], b[[2]]), 1,
                 function(x) c(x[1:m[1]] %x% x[-(1:m[1])])))
  if (ndimx == 3)
    B <- t(apply(cbind(B,  b[[3]]), 1, 
                 function(x) c(x[1:(m[1]*m[2])] %x% x[-(1:(m[1]*m[2]))])))
  
  # Construct smoothness penalty matrices
  P <- list()
  for (i in 1:ndimx) {
    P[[i]] <- diff(diag(m[1]), diff = pord)
    P[[i]] <- t(P[[i]]) %*% P[[i]]
  }
  if (ndimx >= 2) {
    P[[1]] <- P[[1]] %x% diag(m[2])
    P[[2]] <- diag(m[2]) %x% P[[2]]
  }
  if (ndimx == 3) {
    P[[1]] <- P[[1]] %x% diag(m[3])
    P[[2]] <- P[[2]] %x% diag(m[3])
    P[[3]] <- diag(m[1]) %x% diag(m[2]) %x% P[[3]]
  }
  pmat <- matrix(0, nrow = ncol(B), ncol = ncol(B))
  for (i in 1:length(P))
    pmat <- pmat + P[[i]]
  
  #     Construct anova constraint penalty matrices
  if (length(ndims) == 1) {
    # pmat <- pmat + matrix(1, nrow = prod(m), ncol = prod(m))
    Bsum <- apply(B, 2, sum)
    pmat <- pmat + Bsum %o% Bsum
    cmat <- Bsum %o% Bsum
  }
  else if (length(ndims) == 2) {
    if (all(ndims == c(1, 1))) ind <- c(m[1], m[2])
    if (all(ndims == c(1, 2))) ind <- c(m[1], m[2] * m[3])
    if (all(ndims == c(2, 1))) ind <- c(m[1] * m[2], m[3])
    pmat <- pmat + matrix(1, nrow = ind[1], ncol = ind[1]) %x% diag(ind[2])
    pmat <- pmat + diag(ind[1]) %x% matrix(1, nrow = ind[2], ncol = ind[2])
  }
  else if (length(ndims) == 3) {
    pmat <- pmat + matrix(1, nrow = m[1], ncol = m[1]) %x% diag(m[2]) %x% diag(m[3])
    pmat <- pmat + diag(m[1]) %x% matrix(1, nrow = m[2], ncol = m[2]) %x% diag(m[3])
    pmat <- pmat + diag(m[1]) %x% diag(m[2]) %x% matrix(1, nrow = m[3], ncol = m[3])
  }
  
  invisible(list(B = B, P = pmat, cmat = cmat, xrange = xrange, nseg = nseg, bdeg = bdeg, pord = pord))
}

s <- function(..., lambda = NA, df = NA, period = NA, xrange = NA, nseg = NA) {
  vars.list <- as.list(substitute(list(...)))[-1]
  nvar <- length(vars.list)
  if (nvar > 3)
    stop("smooth terms can be constructed from only 1, 2 or 3 variables.")
  if (any(is.na(xrange))) xrange <- matrix(NA, nrow = nvar, ncol = 2)
  variables <- character(0)
  for (i in 1:nvar) variables <- c(variables, deparse(vars.list[[i]]))
  list(variables = variables, lambda = lambda, df = df, period = period, xrange = xrange,
       nseg = nseg)
}

"sm.options" <- function (...) {
  if (nargs() == 0) return(.sm.Options)
  current <- .sm.Options
  if (is.character(...))
    temp <- eval(parse(text = paste(c("list(", ..., ")"))))
  else temp <- list(...)
  if (length(temp) == 1 && is.null(names(temp))) {
    arg <- temp[[1]]
    switch(mode(arg),
           list = temp <- arg,
           character = return(.sm.Options[arg]),
           stop("invalid argument: ", sQuote(arg)))
  }
  if (length(temp) == 0) return(current)
  n <- names(temp)
  if (is.null(n)) stop("options must be given by name")
  changed <- current[n]
  current[n] <- temp
  if (sys.parent() == 0) env <- asNamespace("sm") else env <- parent.frame()
  assign(".sm.Options", current, envir = env)
  invisible(current)
}

"replace.na" <- function (List, comp, value) {
  arg <- paste(substitute(List), "$", substitute(comp), sep = "")
  arg.value <- eval(parse(text = arg), parent.frame(1))
  if (any(is.na(arg.value))) {
    change <- paste(arg, "<-", deparse(substitute(value)))
    a <- eval(parse(text = change), parent.frame(1))
  }
  invisible()
}

"sm.check.data" <- function (x, y = NA, weights = NA, group = NA, ...) {
  opt <- sm.options(list(...))
  
  density <- all(is.na(y))
  if (density) X <- x
  else  X <- cbind(x, y)
  
  if(all(is.na(weights)) | all(weights == 1))
    X <- cbind(X, 1) 
  else{
    if(!is.na(opt$nbins) & opt$nbins!=0) 
      stop("if weights are set, nbins must be either 0 or NA")
    if(any(weights<0 | is.na(weights))) 
      stop("negative or NA weights are meaningless")
    if(any(weights!=round(weights))) {
      weights <- round(weights/min(weights[weights>0]))
      if(opt$verbose>0) 
        cat("Warning: weights have been rescaled to integer values\n")
    }
    X <- cbind(X, weights)
  }
  
  ndim <- ncol(X) - 1 - (!density)            # dimensionality of x
  if (!all(is.na(group))) {
    X <- cbind(X, group)
    group.col <- ncol(X)
  }
  if (!all(is.na(opt$h.weights))) {
    X <- cbind(X,opt$h.weights)
    hw.col <- ncol(X)
  } 
  if (any(is.na(X)) & opt$verbose > 0) cat("missing data are removed\n")
  X <- na.omit(data.matrix(X))
  if (ndim > 2 + density) stop("x has too many columns")
  weights <- as.vector(X[, ndim + (!density) + 1])
  if (!density) y <- as.vector(X[, ndim + 1])
  x <- if (ndim == 1) as.vector(X[, 1]) else X[, 1:ndim]
  if (!all(is.na(group))) group <- as.vector(X[, group.col])
  if (!all(is.na(opt$h.weights))) opt$h.weights <- X[, hw.col]
  list(x = x, y = y, weights = weights, group = group, ndim = ndim, 
       nobs = nrow(X), density = density, options = opt)
}

".sm.Options" <-
  list(hmult = 1, h.weights = NA, period = NA,
       add = FALSE, band = NA, props = c(75, 50, 25), nbins = NA,
       positive = FALSE, delta = NA, display = NA, 
       xlab = NA, ylab = NA, zlab = NA, 
       xlim = NA, ylim = NA, zlim = NA, yht = NA, asp = NA,
       panel = FALSE, panel.plot = NA,
       ngrid = NA, eval.points = NA, rugplot = TRUE, 
       col = NA, col.band = "cyan", col.mesh = "black", 
       col.palette = topo.colors, col.palette.fn = topo.colors, nlevels = 40, col.points = "black",
       se = NA, se.breaks = c(-3, -2, 2, 3), lty = 1, pch = 1, cex = NA,
       theta = -30, phi = 40, size = 2, scaling = NULL, 
       alpha = 0.7, alpha.mesh = 1, lit = FALSE,
       poly.index = 1, diff.ord = 2, test = NA, hull = TRUE, verbose = 1, 
       df = NA, df.max = NA, method = NA, structure.2d = "scaled", nboot = 100, 
       describe = TRUE, show.script = TRUE, eval.grid = TRUE, reference = NA,
       mask.method = "hull", nlevels = 40, partial.residuals = TRUE,
       transform.response = I)





